# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic datatypes for parsed pyc files."""

from dataclasses import dataclass

from typing import Any, List, Optional, Tuple, Union


@dataclass(kw_only=True)
class CodeTypeBase:
    """Pure python types.CodeType with python version added."""

    python_version: Tuple[int, int]
    co_argcount: int
    co_posonlyargcount: int
    co_kwonlyargcount: int
    co_stacksize: int
    co_flags: int
    co_code: bytes
    co_consts: List[object]
    co_names: List[str]
    co_filename: Union[bytes, str]
    co_name: int
    co_firstlineno: int

    def __repr__(self):
        return f"<code: {self.co_name}>"


@dataclass(kw_only=True)
class CodeType38(CodeTypeBase):
    """CodeType for python 3.8 - 3.10."""

    co_nlocals: int
    co_lnotab: bytes
    co_varnames: List[str]
    co_freevars: Tuple[str, ...]
    co_cellvars: Tuple[str, ...]


@dataclass(kw_only=True)
class CodeType311(CodeTypeBase):
    """CodeType for python 3.11+."""

    co_qualname: str
    co_localsplusnames: Tuple[str, ...]
    co_localspluskinds: Tuple[int, ...]
    co_linetable: bytes
    co_exceptiontable: bytes


@dataclass(kw_only=True)
class Opcode:
    """Opcode with names and line numbers."""

    index: int
    line: int
    op: int
    name: str
    arg: Optional[int]
    argval: Any

    def __str__(self):
        if self.arg is not None:
            return f"{self.line:>5}{self.index:>6}  {self.name:<30}{self.arg:>5} {self.argval}"
        else:
            return f"{self.line:>5}{self.index:>6}  {self.name:<30}"



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.bytecode."""

import unittest

from . import base
from pycnite import bytecode
from pycnite import pyc


class TestBytecode(unittest.TestCase):
    """Test bytecode reading."""

    def test_reader(self):
        path = base.test_pyc("trivial", (3, 11))
        code = pyc.load_file(path)
        opcodes = [x.name for x in bytecode.dis(code)]
        expected = [
            "RESUME",
            "LOAD_CONST",
            "STORE_NAME",
            "LOAD_CONST",
            "STORE_NAME",
            "LOAD_NAME",
            "LOAD_NAME",
            "BINARY_OP",
            "STORE_NAME",
            "LOAD_NAME",
            "LOAD_NAME",
            "BINARY_OP",
            "STORE_NAME",
            "LOAD_CONST",
            "RETURN_VALUE",
        ]
        self.assertEqual(opcodes, expected)

    def test_argval(self):
        def run(version):
            path = base.test_pyc("trivial", version)
            code = pyc.load_file(path)
            opcodes = [(x.name, x.argval) for x in bytecode.dis(code)]
            if version == (3, 11):
                opcodes = opcodes[1:5]
            else:
                opcodes = opcodes[0:4]
            expected = [
                ("LOAD_CONST", 1),
                ("STORE_NAME", "x"),
                ("LOAD_CONST", 2),
                ("STORE_NAME", "y"),
            ]
            self.assertEqual(opcodes, expected)

        for version in base.VERSIONS:
            run(version)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.linetable."""

import unittest

from . import base
from pycnite import linetable
from pycnite import pyc


class TestLineTable(unittest.TestCase):
    """Test linetable parsing."""

    def test_read(self):
        src_file = base.test_src("trivial")
        with open(src_file, "r") as f:
            src = f.readlines()
        n_lines = len(src)
        for version in base.VERSIONS:
            path = base.test_pyc("trivial", version)
            code = pyc.load_file(path)
            lt = linetable.linetable_reader(code)
            entries = lt.read_all()
            self.assertEqual(entries[-1].line, n_lines)


if __name__ == "__main__":
    unittest.main()
